/*
Wormlings. LightCycles/Tron clone for Allegro SizeHack 2000 competition
10240 bytes of horribly written code :-)

See the readme file for more info.

Last modified: 18.04.2000

Copyright (C) 2000 Daniel Schlyder

Do what you want with this source (as long as it's non-profit), but give me credit for the original work.
If you improve the code in any way, please send me the modified source.

Thanks.

email: dsc@c2i.net
  web: http://home.c2i.net/dsc/
*/



#include <time.h> // time()
#include <math.h> // cos(), sin()
#include <allegro.h>



#define VER "1.0"

#define W 640 // arena width
#define H 480 // arena height

// colours
#define BLACK makecol(0, 0, 0)
#define WHITE makecol(255, 255, 255)
#define PINK makecol(255, 0, 255)
#define BLUE makecol(0, 0, 255)
#define W_SHADOW makecol(MAX(0, getr(p[i].col) - 64), MAX(0, getg(p[i].col) - 64), MAX(0, getb(p[i].col) - 64))

// these are just defined to save some much needed space
// much easier to define these, than write better code :-)
#define UL unsigned long
#define S screen
#define F font
#define MOVEMENT (int)p[i].x != p[i].head->x || (int)p[i].y != p[i].head->y
#define PPL i = 0; i < ppl; i++
#define RK readkey() >> 8

// also replaced TRUE and FALSE with 1 and 0 to save a few bytes



typedef struct POS {
   int x, y;
   POS *nxt, *pre;
};

struct {
   int dead, dir, col, del, keyL, keyR, keyT, keyS, speed, pwr, score;
   double x, y;
   POS *head, *tail;
} p[4];

struct {
   int draw, col;
} a[W][H];

struct {
   int dir, turn;
   double x, y;
} b[15];



BITMAP *db, *bg;

volatile int latency, sec;

int ppl = 2, slim = 10, bnum = 5, bsize = 8, size = 2, turnR = 3, maxP = 250, pPerSec = 5, del = 1, alive, winner;

double cosT[360], sinT[360];



void incLatency() {
   latency++;
}
END_OF_FUNCTION(incLatency);



void incSec() {
   sec++;
}
END_OF_FUNCTION(incSec);



UL add15(UL x, UL y, UL n) {
   return makecol15(MIN(getr15(x) + getr15(y), 255), MIN(getg15(x) + getg15(y), 255), MIN(getb15(x) + getb15(y), 255));
}



UL add16(UL x, UL y, UL n) {
   return makecol16(MIN(getr16(x) + getr16(y), 255), MIN(getg16(x) + getg16(y), 255), MIN(getb16(x) + getb16(y), 255));
}



UL add24(UL x, UL y, UL n) {
   return makecol24(MIN(getr24(x) + getr24(y), 255), MIN(getg24(x) + getg24(y), 255), MIN(getb24(x) + getb24(y), 255));
}



void init() {
   int i, x, y;
   
   allegro_init();
   
   install_timer();
   LOCK_VARIABLE(latency);
   LOCK_FUNCTION(incLatency);
   LOCK_VARIABLE(sec);
   LOCK_FUNCTION(incSec);
   
   install_keyboard();
   text_mode(-1);
   
   override_config_file("wormlings.cfg");
   
   srand(time(NULL));

   set_color_depth(get_config_int(NULL, "bpp", 16));
   if (set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0)) {
      allegro_message("%s", allegro_error);
      exit(1);
   }

   set_blender_mode(add15, add16, add24, 0, 0, 0, 0);

   db = create_bitmap(W, H);
   
   bg = create_bitmap(W, H);
   clear(bg);
   for (x = 0; x < W; x += 32)
      for (y = 0; y < H; y += 32) {
         rect(bg, 7 + x, 7 + y, 24 + x, 24 + y, makecol(0, 0, 255 - x / 16 - y / 16));
         rectfill(bg, 8 + x, 8 + y, 23 + x, 23 + y, makecol(0, 0, 128 - x / 16 - y / 16));
      }

   // look-up tables
   for (i = 0; i < 360; i++) {
      cosT[i] = cos(i * 3.141593 / 180);
      sinT[i] = sin(i * 3.141593 / 180);
   }

   p[0].col = makecol(255, 0, 0);
   p[1].col = makecol(0, 255, 0);
   p[2].col = makecol(255, 255, 0);
   p[3].col = makecol(0, 255, 255);
   
   for (i = 0; i < 4; i++)
      p[i].head = p[i].tail = NULL;
}



void end() {
   destroy_bitmap(db);
   destroy_bitmap(bg);
}



// textprintfCentreShadow
void tpcs(BITMAP *b, FONT *f, int x, int y, int c, char *fmt, ...) {
   char buf[512];

   va_list ap;
   va_start(ap, fmt);
   vsprintf(buf, fmt, ap);
   va_end(ap);

   textout_centre(b, f, buf, x + 1, y + 1, BLACK);
   textout_centre(b, f, buf, x, y, c);
}



// screenshot
void ss() {
   static int ss_c = 0;
   char fname[15];
	BITMAP *b;

   sprintf(fname, "ss%03d.tga", ++ss_c);
   while (exists(fname))
      sprintf(fname, "ss%03d.tga", ++ss_c);

	b = create_sub_bitmap(S, 0, 0, W, H);

   save_tga(fname, b, NULL);

	destroy_bitmap(b);
}



void frame() {
   int x, y;
   
   // draw arena
   for (y = 0; y < H; y++)
      for (x = 0; x < W; x++)
         if (a[x][y].draw) {
            if (a[x][y].col)
               putpixel(db, x, y, a[x][y].col);
            else
               putpixel(db, x, y, getpixel(bg, x, y));
            a[x][y].draw = 0; // FALSE
         }

   // display players power status
   if (maxP) {
      blit(bg, db, W - 80, H - 8, W - 80, H - 8, 80, 8);
      textprintf(db, F, W - 80, H - 8, p[0].col, "Power: %03d", p[0].pwr);
      blit(bg, db, 0, 0, 0, 0, 80, 8);
      textprintf(db, F, 0, 0, p[1].col, "Power: %03d", p[1].pwr);
      if (ppl > 2) {
         blit(bg, db, 0, H - 8, 0, H - 8, 80, 8);
         textprintf(db, F, 0, H - 8, p[2].col, "Power: %03d", p[2].pwr);
      }
      if (ppl == 4) {
         blit(bg, db, W - 80, 0, W - 80, 0, 80, 8);
         textprintf(db, F, W - 80, 0, p[3].col, "Power: %03d", p[3].pwr);
      }
   }

   blit(db, S, 0, 0, 0, 0, W, H);
}



// keyboardInput
void kIn() {
   int i, l, s;
   
   for (PPL) {
      if (!p[i].dead) {
         if (key[p[i].keyL]) {
            p[i].dir += turnR;
            if (p[i].dir > 359)
               p[i].dir = 0;
         }
         else if (key[p[i].keyR]) {
            p[i].dir -= turnR;
            if (p[i].dir < 0)
               p[i].dir = 359;
         }
         p[i].speed = 1;
         if (maxP && p[i].pwr) {
            if (key[p[i].keyS]) {
               p[i].speed = 0;
               p[i].pwr--;
            }
            else if (key[p[i].keyT]) {
               p[i].speed = 2;
               p[i].pwr--;
            }
         }
      }
   }

   if (key[KEY_ESC])
      winner = -2;
      
   if (key[KEY_F12]) {
      l = latency;
      s = sec;
      ss();
      while (key[KEY_F12]) {}
      latency = l;
      sec = s;
   }
}



// doCircleFill
void doCF(BITMAP *b, int x, int y, int r, int d, void (*proc)(BITMAP *, int, int, int)) {
   int cx = 0, cy = r, df = 1 - r, d_e = 3, d_se = -2 * r + 5;

   do {
      do_line(b, x - cy, y - cx, x + cy, y - cx, d, proc);

      if (cx)
         do_line(b, x - cy, y + cx, x + cy, y + cx, d, proc);

      if (df < 0) {
       	df += d_e;
         d_e += 2;
         d_se += 2;
      }
      else { 
	      if (cx != cy) {
	         do_line(b, x - cx, y - cy, x + cx, y - cy, d, proc);

   	      if (cy)
   	         do_line(b, x - cx, y + cy, x + cx, y + cy, d, proc);
	      }

   	   df += d_se;
   	   d_e += 2;
   	   d_se += 4;
   	   cy--;
      } 

      cx++;

   } while (cx <= cy);
}



// wrapCoords
template<class T>
void wC(T& x, T& y) {
   if (x < 0)
      x += W;
   else if (x >= W)
      x -= W;
   if (y < 0)
      y += H;
   else if (y >= H)
      y -= H;
}



void collW1(BITMAP *b, int x, int y, int i) {
   wC(x, y);

   if (a[x][y].col && a[x][y].col != p[i].col)
      p[i].dead = 1; // TRUE
}



void collW2(BITMAP *b, int x, int y, int i) {
   wC(x, y);

   if (a[x][y].col && a[x][y].col != p[i].col && a[x][y].col != W_SHADOW)
      p[i].dead = 1; // TRUE
}



void collB(BITMAP *bp, int x, int y, int i) {
   if (x < 0 || x >= W || y < 0 || y >= H)
      b[i].turn = 1; // TRUE
   else if (a[x][y].col)
      b[i].turn = 1; // TRUE
}



void plot(BITMAP *b, int x, int y, int col) {
   wC(x, y);

   a[x][y].col = col;
   a[x][y].draw = 1; // TRUE
}



void plotIfClear(BITMAP *b, int x, int y, int col) {
   wC(x, y);

   if (!a[x][y].col) {
      a[x][y].col = col;
      a[x][y].draw = 1; // TRUE
   }
}



void clearIfNotCol(BITMAP *b, int x, int y, int col) {
   wC(x, y);

   if (a[x][y].col != col) {
      a[x][y].col = 0;
      a[x][y].draw = 1; // TRUE
   }
}



void newH(int i) {
   POS *tmp = new POS;

   if (p[i].head)
      p[i].head->pre = tmp;
   tmp->nxt = p[i].head;
   tmp->pre = NULL;
   tmp->x = (int)p[i].x;
   tmp->y = (int)p[i].y;
   p[i].head = tmp;
}



void delT(int i) {
   POS *tmp = p[i].tail;
   
   if (tmp) {
      
      p[i].tail = tmp->pre;

      if (p[i].tail) {
         p[i].tail->nxt = NULL;
         
         // erase old tail
         doCF(db, tmp->x, tmp->y, size, p[i].col, clearIfNotCol);
         doCF(db, tmp->x, tmp->y, size - 1, 0, plot);

         // draw new tail
         doCF(db, p[i].tail->x, p[i].tail->y, size - 1, p[i].col, plot);
         doCF(db, p[i].tail->x, p[i].tail->y, size, W_SHADOW, plotIfClear);
      }
      else {
         p[i].head = NULL;
         doCF(db, tmp->x, tmp->y, size, 0, plot);
      }
      
      delete tmp;
   }
}



// newRound
void nr() {
   int x, y, i;
   
   // clear arena
   for (x = 0; x < W; x++)
      for (y = 0; y < H; y++) {
         a[x][y].col = 0;
         a[x][y].draw = 0; // FALSE
      }
   
   // blit background to buffer
   blit(bg, db, 0, 0, 0, 0, W, H);

   // init worms
   p[0].dir = 135;
   p[0].x = W - 9;
   p[0].y = H - 9;

   p[1].dir = 315;
   p[1].x = 8;
   p[1].y = 8;

   p[2].dir = 45;
   p[2].x = 8;
   p[2].y = H - 9;

   p[3].dir = 225;
   p[3].x = W - 9;
   p[3].y = 8;

   for (PPL) {
    	while (p[i].head) {
   	   p[i].tail = p[i].head->nxt;
   	   delete p[i].head;
   	   p[i].head = p[i].tail;
   	}

      newH(i);
      p[i].tail = p[i].head;

      p[i].dead = 0; // FALSE
      p[i].del = 0; // FALSE
      p[i].pwr = maxP;
   }

   alive = ppl;
   
   // init balls
   if (bnum) {
      for (i = 0; i < bnum; i++) {
         b[i].dir = rand() % 360;
         do {
            b[i].x = rand() % (W - 80) + 40;
            b[i].y = rand() % (H - 60) + 30;
            b[i].turn = 0; // FALSE
            do_circle(db, (int)b[i].x, (int)b[i].y, bsize, i, collB);
         } while (b[i].turn);
      }
   }
   
   // reset timer variables
   latency = 0;
   sec = 0;
}



void updateB() {
   int i;
   double x, y;
   
   for (i = 0; i < bnum; i++) {
      x = b[i].x;
      y = b[i].y;
      b[i].x += cosT[b[i].dir];
      b[i].y -= sinT[b[i].dir];

      if ((int)b[i].x != (int)x || (int)b[i].y != (int)y) {
         doCF(db, (int)x, (int)y, bsize, 0, plot); // erase old ball position
         b[i].turn = 0; // FALSE
         doCF(db, (int)b[i].x, (int)b[i].y, bsize, i, collB); // collision check
         if (b[i].turn) {
            b[i].dir = rand() % 360;
            b[i].x = x;
            b[i].y = y;
         }
         // draw new ball position
         do_circle(db, (int)b[i].x, (int)b[i].y, bsize, makecol(128, 0, 128), plot);
         doCF(db, (int)b[i].x, (int)b[i].y, bsize, makecol(192, 0, 192), plotIfClear);
      }
   }
}



void updateW() {
   int i, t, s, pwrInc = sec;
   POS *tmp;
   
   // reset time counter
   sec = 0;

   for (PPL) {
      if (!p[i].dead) {
         // increase power per second elapsed
         p[i].pwr = MIN(p[i].pwr + pwrInc * pPerSec, maxP);

         if (p[i].speed) {
            p[i].x += cosT[p[i].dir];
            p[i].y -= sinT[p[i].dir];
   
            wC(p[i].x, p[i].y);
   
            if (MOVEMENT) {
               if (p[i].del)
                  delT(i);
               p[i].del = !p[i].del;
            }
         }
      }
      else
         if (del)
            delT(i);
   }

   for (s = 0; s < 2; s++) {
      for (PPL) {
         if (!p[i].dead) {
            if (s && p[i].speed == 2) {
               p[i].x += cosT[p[i].dir];
               p[i].y -= sinT[p[i].dir];
   
               wC(p[i].x, p[i].y);
            }
   
            if (MOVEMENT) {
               // erase a few previous heads' borders to avoid collision detected vs these
               tmp = p[i].head;
               for (t = 0; t <= size; t++) {
                  if (tmp) {
                     doCF(db, tmp->x, tmp->y, size, p[i].col, clearIfNotCol);
                     tmp = tmp->nxt;
                  }
                  else
                     break;
               }
               
               // collision check vs old situation
               do_circle(db, (int)p[i].x, (int)p[i].y, size, i, collW1);
               
               // redraw old heads' borders
               tmp = p[i].head;
               for (t = 0; t <= size; t++) {
                  if (tmp) {
                     doCF(db, tmp->x, tmp->y, size, W_SHADOW, plotIfClear);
                     tmp = tmp->nxt;
                  }
                  else
                     break;
               }
   
               // draw new head
               doCF(db, (int)p[i].x, (int)p[i].y, size - 1, p[i].col, plot);
               doCF(db, (int)p[i].x, (int)p[i].y, size, W_SHADOW, plotIfClear);
   
               newH(i);
            }
         }
      }
   
      // detect collisions vs new situation
      for (PPL)
         if (!p[i].dead)
            if (MOVEMENT)
               do_circle(db, (int)p[i].x, (int)p[i].y, size, i, collW2);
   }
}



void updateS() {
   int i, a = 0, smax = 0;
   
   // find players alive
   for (PPL)
      if (!p[i].dead)
         a++;
   
   // if # changed: award difference to players still alive
   if (a < alive) {
      for (PPL)
         if (!p[i].dead)
            p[i].score += alive - a;
   }
   
   // check if score limit reached. if so, find winner(s)
   if (a <= 1) {
      for (PPL) {
         smax = MAX(p[i].score, smax);
         if (smax >= slim) {
            for (PPL)
               if (p[i].score == smax) {
                  if (!winner)
                     winner = i + 1;
                  else
                     winner = -1; // game ended in a tie
               }
         }
      }
   }
      
   alive = a;
}



void displayS() {
   BITMAP *b = create_bitmap(8 * 20, 16 * ppl + 8);
   int i, smax = 0;

   clear_to_color(b, BLUE);
   set_trans_blender(0, 0, 0, 128);
   draw_trans_sprite(db, b, W / 2 - b->w / 2, H / 2 - b->h / 2);
   destroy_bitmap(b);

   rect(db, W / 2 - b->w / 2 - 1, H / 2 - b->h / 2 - 1, W / 2 + b->w / 2, H / 2 + b->h / 2, BLUE);

   if (winner) {
      tpcs(db, F, W / 2, H / 2 - b->h / 2 + 8, WHITE, "Game Over");
      if (winner > 0)
         tpcs(db, F, W / 2, H / 2 + b->h / 2 - 16, p[winner - 1].col, "Player %d won!", winner);
      else
         tpcs(db, F, W / 2, H / 2 + b->h / 2 - 16, WHITE, "A tie. How boring!");
   }
   else {
      for (PPL)
         smax = MAX(p[i].score, smax);
      for (PPL) {
         tpcs(db, F, W / 2, H / 2 - b->h / 2 + 8 + 16 * i, p[i].col, "Player %d: %2d", i + 1, p[i].score);
         if (p[i].score == smax) {
            tpcs(db, F, W / 2 - b->w / 2 + 16, H / 2 - b->h / 2 + 8 + 16 * i, WHITE, ">");
            tpcs(db, F, W / 2 + b->w / 2 - 16, H / 2 - b->h / 2 + 8 + 16 * i, WHITE, "<");
         }
      }
   }

   blit(db, S, 0, 0, 0, 0, W, H);

   rest(250);
   clear_keybuf();
   readkey();
}



int title() {
   BITMAP *b = create_bitmap(72, 8);
   int y;

   clear(db);
   for (y = 0; y < H; y += 2)
      hline(db, 0, y, W, makecol((H - y) * 128 / H, 0, (H - y) * 128 / H));

   clear_to_color(b, PINK);
   textout(b, F, "Wormlings", 0, 0, BLACK);
   stretch_sprite(db, b, 12, H / 2, W, H / 2);
   textout(b, F, "Wormlings", 0, 0, makecol(0, 0, 128));
   stretch_sprite(db, b, 4, H / 2 - 8, W, H / 2);
   destroy_bitmap(b);

   tpcs(db, F, W / 2, H - 9, BLUE, "Version %s (C) 2000 Daniel Schlyder", VER);

   blit(db, S, 0, 0, 0, 0, W, H);

   clear_keybuf();
   if (RK == KEY_ESC)
      return 0; // FALSE

   return 1; // TRUE
}



// setValue
void setV(int& val, int min, int max, int step, int y, char *msg) {
   int fi = 0; // FALSE
   char string[25];
   
   textout(db, F, msg, W / 2 - text_length(F, msg), y, BLUE);
   clear_keybuf();
   
   do {
      rectfill(db, W / 2 + 7, y - 1, W / 2 + 7 + 8 * strlen(itoa(max, string, 10)), y + 7, BLACK);
      textprintf(db, F, W / 2 + 8, y, WHITE, "%0*d", strlen(itoa(max, string, 10)), val);
      blit(db, S, 0, 0, 0, 0, W, H);
      if (keypressed()) {
         switch (RK) {
            case KEY_LEFT:
               val = MAX(val - step, min);
               break;
            case KEY_RIGHT:
               val = MIN(val + step, max);
               break;
            case KEY_ENTER:
               fi = 1; // TRUE
               break;
         }
      }
   } while (!fi);
}



void options() {
   int i, y = 0;
   
   setV(ppl, 2, 4, 1, 16, "# of Players:");
   setV(size, 1, 3, 1, 32, "Player Size:");
   setV(turnR, 1, 5, 1, 48, "Player Turn Radius (higher is better):");
   setV(maxP, 0, 500, 5, 64, "Max Power:");
   if (maxP) {
      setV(pPerSec, 0, 10, 1, 80, "Power Regained Each Second:");
      y += 16;
   }
   setV(del, 0, 1, 1, 80 + y, "Erase the Dead:");
   setV(bnum, 0, 15, 1, 96 + y, "# of Deathballs:");
   if (bnum) {
      setV(bsize, 2, 10, 1, 112 + y, "Ball Size:");
      y += 16;
   }
   setV(slim, 1, 99, 1, 112 + y, "Score Limit:");

   for (PPL) {
      tpcs(S, F, W / 2, 128 + y + 44 * i, p[i].col, "Player %d's TURN LEFT key", i + 1);
      p[i].keyL = RK;
      tpcs(S, F, W / 2, 128 + y + 44 * i + 8, p[i].col, "Player %d's TURN RIGHT key", i + 1);
      p[i].keyR = RK;
      tpcs(S, F, W / 2, 128 + y + 44 * i + 20, p[i].col, "Player %d's TURBO key", i + 1);
      p[i].keyT = RK;
      tpcs(S, F, W / 2, 128 + y + 44 * i + 28, p[i].col, "Player %d's STOP key", i + 1);
      p[i].keyS = RK;
   }
   
   tpcs(S, F, W / 2, 144 + y + 44 * i, WHITE, "Game starts in 2 seconds...");
   rest(2000);
}



// newGame
void ng() {
   int i;
   
   for (PPL)
      p[i].score = 0;
   
   winner = 0;

   install_int_ex(incLatency, MSEC_TO_TIMER(10));
   install_int_ex(incSec, SECS_TO_TIMER(1));
}



int main() {
   init();

   while (title()) {
      options();
      ng();
      nr();
      
      while (!winner) {
     	   while (latency > 0 && !winner) {
     	      kIn();
     	      if (bnum)
     	         updateB();
     	      updateW();
     	      updateS();

            if (--latency == 0)
               frame();

        	   if (alive <= 1) {
        	      frame();
     	         displayS();

     	         if (!winner)
        	         nr();
        	   }
         }
      }

      remove_int(incLatency);
      remove_int(incSec);
   }

   end();
   
   return 0;
}
END_OF_MAIN();
